/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Computes the interface identifier for a given header defined
 * by the pseudorandom approach suggested in "vrpsx_identified.h".
 */

#define _POSIX_C_SOURCE 200809L

#include "vrpsx_identify_lib.h"
#include <vrpsx_ostreambuf_default.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>


#define MALLOC(TYPE)  ((TYPE *)malloc(sizeof(TYPE)))


typedef struct FILE_ostreambuf_struct {
    vrpsx_ostreambuf_t interface;
    FILE *fp;
    char buffer[64];
} FILE_ostreambuf_t;

static vrpsx_ostreambuf_vtbl_t const FILE_ostreambuf_vtbl;

static int FILE_ostreambuf_create(
    FILE *fp,
    vrpsx_ostreambuf_t **osb)
{
    *osb = NULL;

    FILE_ostreambuf_t *obj = MALLOC(FILE_ostreambuf_t);
    if (! obj) {
        return errno;
    }

    obj->interface.vtbl = &FILE_ostreambuf_vtbl;
    obj->interface.current = obj->buffer;
    obj->interface.end = obj->buffer + sizeof obj->buffer;
    obj->interface.error = 0;
    obj->fp = fp;

    *osb = &obj->interface;
    return 0;
}

static int FILE_ostreambuf_disown(
    vrpsx_ostreambuf_t **osb)
{
    vrpsx_ostreambuf_t *doomed = *osb;
    *osb = NULL;
    int ern = VRPSX_CALL0(doomed, flush);
    VRPSX_UPCAST_NONNULL_DECL(FILE_ostreambuf_t, , obj, interface, doomed);
    free(obj);
    return ern;
}

static int FILE_ostreambuf_flush(
    vrpsx_ostreambuf_t *osb)
{
    VRPSX_UPCAST_NONNULL_DECL(FILE_ostreambuf_t, , obj, interface, osb);
    size_t buffered = obj->interface.current - obj->buffer;

    if (obj->interface.error) {
        return obj->interface.error;
    }

    if (fwrite(obj->buffer, 1, buffered, obj->fp) != buffered) {
        obj->interface.error = errno;
        return obj->interface.error;
    }

    obj->interface.current = obj->buffer;
    return 0;
}

static vrpsx_ostreambuf_end_t const vrpsx_ostreambuf_end;

static vrpsx_ostreambuf_vtbl_t const FILE_ostreambuf_vtbl = {
    &FILE_ostreambuf_disown,
    &FILE_ostreambuf_flush,
    &vrpsx_ostreambuf_put1_from_flush,
    &vrpsx_ostreambuf_putn_from_put1,
    &vrpsx_ostreambuf_end
};


static void usage(
    char const *argv0)
{
    fprintf(stderr,
            "Usage: %s \\\n"
            " [-o <out-file>] [-h <out-header> -m <macro-name>] \\\n"
            " [-n <display-name>] [-d <debug-out-file>] \\\n"
            " [-I <include-dir>] <in-header> <name-within-header>\n"
            "\n"
            "This program computes the interface identifier for\n"
            "the interface defined by the specified input header\n"
            "that has the specified name.\n"
            "\n"
            "-o: Write the identifier to the specified file, followed\n"
            "    by a newline; '-' and omission mean standard output.\n"
            "\n"
            "-h: Write a C/C++ header with this name;\n"
            "    '-' means standard output.  Use with '-m'.\n"
            "\n"
            "-m: The base name of the macros defined in that header;\n"
            "    for example, the one ending in '_IDENTIFIER' holds the\n"
            "    numeric value of the computed interface identifier.\n"
            "    Use with '-h'.\n"
            "\n"
            "-n: Specifies the display name of the interface in question,\n"
            "    which need not uniquely identify that interface.\n"
            "    Defaults to the <name-within-header> argument.\n"
            "\n"
            "-d: Write to this debug output file a copy of the input to\n"
            "    the MD5 digest algorithm; '-' means standard output.\n"
            "\n"
            "-I: Lookup headers in the specified directory.\n"
            "    Specify multiple times for multiple include directories.\n"
            "\n",
            argv0);
}


int main(
    int argc,
    char **argv)
{
    char *error_message = NULL;
    char const **include_directories = NULL;
    char const **incdir_write = NULL;
    char const *out_file = "-";
    char const *out_header = NULL;
    char const *macro_base = NULL;
    char const *display_name = NULL;
    char const *in_header = NULL;
    char const *name_within_header = NULL;
    char const *debug_path = NULL;
    FILE *debug_fp = NULL;
    FILE *out_file_fp = NULL;
    FILE *out_header_fp = NULL;
    vrpsx_ostreambuf_t *debug_osb = NULL;
    vrpsx_uint32_t iid = 0;
    int ch, ern;
    char iid_str[9];

    include_directories = (char const **)calloc(argc, sizeof(char const*));
    if (! include_directories) {
        fprintf(stderr, "%s: %s\n", argv[0], strerror(errno));
        return 1;
    }
    incdir_write = include_directories;

    ern = 0;

    while ((ch = getopt(argc, argv, ":o:h:m:n:I:d:")) != -1) {
        switch (ch) {
            case 'o': {
                out_file = optarg;
                break;
            }
            case 'h': {
                out_header = optarg;
                break;
            }
            case 'm': {
                macro_base = optarg;
                break;
            }
            case 'n': {
                display_name = optarg;
                break;
            }
            case 'I': {
                *incdir_write++ = optarg;
                break;
            }
            case 'd': {
                debug_path = optarg;
                break;
            }
            case '?': {
                fprintf(stderr, "Unrecognized option: -%c\n", optopt);
                ern = EINVAL;
                break;
            }
        }
    }

    if (optind == argc - 2) {
        in_header = argv[optind];
        name_within_header = argv[optind + 1];
    } else {
        ern = EINVAL;
    }

    if (out_header && ! macro_base) {
        fprintf(stderr, "-h requires -m\n");
        ern = EINVAL;
    }

    if (! display_name) {
        display_name = name_within_header;
    }

    if (ern) {
        usage(argv[0]);
        free(include_directories);
        return 2;
    }

    if (debug_path) {
        debug_fp = strcmp(debug_path, "-") ? fopen(debug_path, "w") : stdout;
        if (! debug_fp) {
            fprintf(stderr, "%s: %s: %s\n",
                    argv[0], debug_path, strerror(errno));
            free(include_directories);
            return 1;
        }

        ern = FILE_ostreambuf_create(debug_fp, &debug_osb);
        if (ern) {
            fprintf(stderr, "%s: %s\n", argv[0], strerror(ern));
            if (debug_fp && debug_fp != stdout) {
                fclose(debug_fp);
            }
            free(include_directories);
        }
    }

    ern = interface_identifier(
        &error_message,
        include_directories,
        in_header,
        name_within_header,
        &iid,
        debug_osb);

    free(include_directories);
    include_directories = NULL;
    if (ern) {
        fprintf(stderr, "%s:\n%s\n", argv[0], error_message);
        free(error_message);
        VRPSX_SINGLE_OWNER_CLEAR(vrpsx_ostreambuf_t, &debug_osb);
        if (debug_fp && debug_fp != stdout) {
            fclose(debug_fp);
        }
        return 1;
    }

    if (debug_osb) {
        VRPSX_SINGLE_OWNER_CLOSE(vrpsx_ostreambuf_t, &debug_osb, &ern);
        if (ern) {
            fprintf(stderr, "%s: %s: %s\n",
                    argv[0], debug_path, strerror(ern));
            if (debug_fp != stdout) {
                fclose(debug_fp);
            }
            return 1;
        }
    }

    if (debug_fp && debug_fp != stdout && fclose(debug_fp)) {
        fprintf(stderr, "%s: %s: %s\n", argv[0], debug_path, strerror(errno));
        return 1;
    }
    debug_fp = NULL;

    interface_identifier_to_string(iid, iid_str);

    /* Write interface identifier, followed by newline.
     */
    out_file_fp = strcmp(out_file, "-") ? fopen(out_file, "w") : stdout;
    if (! out_file_fp) {
        fprintf(stderr, "%s: %s: %s\n",
                argv[0], out_file, strerror(errno));
        return 1;
    }
    if (fwrite(iid_str, 1, 8, out_file_fp) != 8
            || fputc('\n', out_file_fp) == EOF) {
        fprintf(stderr, "%s: %s: %s\n",
                argv[0], out_file, strerror(errno));
        fclose(out_file_fp);
        return 1;
    }
    if (out_file_fp != stdout && fclose(out_file_fp)) {
        fprintf(stderr, "%s: %s: %s\n",
                argv[0], out_file, strerror(errno));
        return 1;
    }
    out_file_fp = NULL;

    /* If requested, write header defining macro for interface identifier.
     */
    if (out_header) {
        out_header_fp = strcmp(out_header, "-")
                ? fopen(out_header, "w") : stdout;
        if (! out_header_fp) {
            fprintf(stderr, "%s: %s: %s\n",
                    argv[0], out_header, strerror(errno));
            return 1;
        }
        if (fprintf(out_header_fp,
                        "#ifndef %s_IDENTIFIER\n"
                        "#define %s_IDENTIFIER 0x%sU\n"
                        "#define %s_IDENTIFIER_RAW %s\n"
                        "#define %s_IDENTIFIER_STR \"%s\"\n"
                        "#define %s_IDENTIFIER_CAT_UNEVALUATED(PREFIX, SUFFIX)"
                        " \\\n        PREFIX ## %s ## SUFFIX\n"
                        "#define %s_IDENTIFIER_CAT(PREFIX, SUFFIX) \\\n    "
                        "    %s_IDENTIFIER_CAT_UNEVALUATED(PREFIX, SUFFIX)\n"
                        "#define %s_DISPLAY_NAME \"%s\"\n"
                        "#endif /* %s_IDENTIFIER */\n",
                        macro_base,
                        macro_base, iid_str,
                        macro_base, iid_str,
                        macro_base, iid_str,
                        macro_base, iid_str,
                        macro_base, macro_base,
                        macro_base, display_name,
                        macro_base) < 0) {
            fprintf(stderr, "%s: %s: %s\n",
                    argv[0], out_header, strerror(errno));
            fclose(out_header_fp);
            return 1;
        }
        if (out_header_fp != stdout && fclose(out_header_fp)) {
            fprintf(stderr, "%s: %s: %s\n",
                    argv[0], out_header, strerror(errno));
            return 1;
        }
        out_header_fp = NULL;
    }

    return 0;
}
